from typing import Generic, TypeVar

from rainbowadn.collection.linear import *
from rainbowadn.core import *
from rainbowadn.flow.core import *
from rainbowadn.flow.sequence import *
from ._listbridge import *

__all__ = ('StackBridge',)

ElementT = TypeVar('ElementT')
Out = TypeVar('Out')


class StackBridge(
    Reducer[SequenceDispatcher[Stack[ElementT], Out], Out],
    Generic[Out, ElementT]
):
    def __init__(self, target: HashPoint[Stack[ElementT]]):
        assert isinstance(target, HashPoint)
        self.target = target

    async def reduce(self, reduce: Reduce[SequenceDispatcher[Stack[ElementT], Out], Out]) -> Out:
        resolved: Stack[ElementT] = await self.target.resolve()
        dispatchers: list[SequenceDispatcher[Stack[ElementT], Out]] = [LastDispatcher(resolved)]
        deepest: Stack[ElementT] = resolved
        stack: Stack[ElementT]
        for stack in [x async for x in Stack.iter_stacks(resolved.previous)]:
            assert isinstance(stack, Stack)
            dispatchers.append(PairDispatcher(stack, deepest))
            deepest = stack
        dispatchers.append(FirstDispatcher(deepest))
        bridge: Reducer[SequenceDispatcher[Stack[ElementT], Out], Out] = ListBridge(dispatchers)
        assert isinstance(bridge, Reducer)
        return await bridge.reduce(reduce)

    def loose(self) -> Reducer[SequenceDispatcher[Stack[ElementT], Out], Out]:
        return self

    @classmethod
    def mapper(cls) -> Mapper[
        HashPoint[Stack[ElementT]],
        Reducer[SequenceDispatcher[Stack[ElementT], Out], Out]
    ]:
        return CallableMapper(cls)

    @classmethod
    def element_of(cls, stack: Stack[ElementT]) -> HashPoint[ElementT]:
        return stack.element

    @classmethod
    def element_mapper(cls) -> Mapper[Stack[ElementT], HashPoint[ElementT]]:
        return CallableMapper(cls.element_of)

    def over_elements(self) -> Reducer[SequenceDispatcher[HashPoint[ElementT], Out], Out]:
        cdm: Mapper[
            SequenceDispatcher[Stack[ElementT], None],
            SequenceDispatcher[HashPoint[ElementT], None]
        ] = CompositionDispatcher.mapper(self.element_mapper())
        return MapReducer(cdm, self)
