from typing import Generic, TypeVar

from rainbowadn.core import *
from rainbowadn.flow.core import *
from rainbowadn.flow.sequence import *
from rainbowadn.flow.verification.core import *
from rainbowadn.nullability import *
from ._stageverificationdispatch import *
from ._stageverificationprotocol import *

__all__ = ('StageVerification',)

Base = TypeVar('Base')
Stage = TypeVar('Stage')
Header = TypeVar('Header')
State = TypeVar('State')


class StageVerification(
    Verification[tuple[Nullable[State], Header, State]],
    Generic[Base, Stage, Header, State]
):
    def __init__(
            self,
            protocol: StageVerificationProtocol[Base, Stage, Header],
            mapper: Mapper[State, tuple[Base, Reducer[SequenceDispatcher[Stage, bool], bool]]],
    ):
        assert isinstance(protocol, StageVerificationProtocol)
        assert isinstance(mapper, Mapper)
        self.protocol = protocol
        self.mapper = mapper

    async def _base(self, state: Nullable[State]) -> Nullable[State]:
        assert isinstance(state, Nullable)
        if state.null():
            return Null()
        else:
            base: Base
            base, _ = await self.mapper.map(state.resolve())
            return NotNull(base)

    async def verify(self, element: tuple[Nullable[State], Header, State]) -> bool:
        assert isinstance(element, tuple)
        previous: Nullable[State]
        header: Header
        state: State
        previous, header, state = element
        assert isinstance(previous, Nullable)
        previous_base: Nullable[State]
        base: Base
        reducer: Reducer[SequenceDispatcher[Stage, bool], bool]
        previous_base, (base, reducer) = await gather(
            self._base(previous),
            self.mapper.map(state),
        )
        assert isinstance(previous_base, Nullable)
        assert isinstance(reducer, Reducer)
        verification: Verification[Reducer[SequenceDispatcher[Stage, bool], bool]] = StageVerificationDispatch(
            self.protocol,
            previous_base,
            header,
            base
        ).reduce_verification()
        assert isinstance(verification, Verification)
        assert_true(
            await verification.verify(reducer)
        )
        return True

    def loose(self) -> Verification[tuple[Nullable[State], Header, State]]:
        return self
