from typing import Generic, Iterable, TypeVar

from rainbowadn.flow.core import *

__all__ = ('FlowIterate',)

Element = TypeVar('Element')


class FlowIterate(
    Reduce[Element, Iterable[Element]],
    Generic[Element]
):
    async def reduce(self, out: Iterable[Element], element: Element) -> Iterable[Element]:
        def wrap() -> Iterable[Element]:
            yield from out
            yield element

        return wrap()

    def merge(self, left: Iterable[Element], right: Iterable[Element]) -> Iterable[Element]:
        def wrap() -> Iterable[Element]:
            yield from left
            yield from right

        return wrap()

    def loose(self) -> Reduce[Element, Iterable[Element]]:
        return self

    @classmethod
    async def iterate(
            cls,
            reducer: Reducer[Element, Iterable[Element]],
            initial: Iterable[Element]
    ) -> Iterable[Element]:
        return await reducer.reduce(cls(initial))
